import Foundation
import SQLite
import CoreLocation
import HealthKit

typealias Database = Connection

final class HealthDatabase: ObservableObject {
    
    private let fileUrl: URL

    private let database: Connection

    private let samples: SamplesTable

    private let workoutsTable: WorkoutsTable

    private let locationSamples: LocationSeriesDataTable

    private let dataSeries: DataSeriesTable

    @Published
    var workouts: [Workout] = []
    
    convenience init(fileUrl: URL) throws {
        let database = try Connection(fileUrl.path)
        self.init(fileUrl: fileUrl, database: database)
    }

    init(fileUrl: URL, database: Connection) {
        self.fileUrl = fileUrl
        self.database = database
        self.samples = .init(database: database)
        self.workoutsTable = .init(database: database)
        self.locationSamples = .init(database: database)
        self.dataSeries = .init(database: database)

        DispatchQueue.global().async {
            self.readAllWorkouts()
        }
    }

    func readAllWorkouts() {
        let workouts: [Workout]
        do {
            workouts = try workoutsTable.workouts()
        } catch {
            print("Failed to read workouts: \(error)")
            return
        }
        DispatchQueue.main.async {
            self.workouts = workouts
        }
    }

    func locationSamples(for activity: HKWorkoutActivity) throws -> [LocationSample] {
        try locationSamples.locationSamples(from: activity.startDate, to: activity.currentEndDate)
    }

    func locationSampleCount(for activity: HKWorkoutActivity) throws -> Int {
        try locationSamples.locationSampleCount(from: activity.startDate, to: activity.currentEndDate)
    }

    func samples(for activity: HKWorkoutActivity) throws -> [Sample.DataType : [Sample]] {
        try samples.samples(from: activity.startDate, to: activity.currentEndDate).reduce(into: [:]) {
            $0[$1.dataType] = ($0[$1.dataType] ?? []) + [$1]
        }
    }

    func sampleCount(for activity: HKWorkoutActivity) throws -> Int {
        try samples.sampleCount(from: activity.startDate, to: activity.currentEndDate)
    }

    var activities: [HKWorkoutActivity] {
        workouts.map { $0.activities }.joined().sorted()
    }

    private func testActivityOverlap() {
        let activities = self.activities
        var current = activities.first!
        for next in activities.dropFirst() {
            let overlap = next.startDate.timeIntervalSince(current.currentEndDate)
            if overlap < 0 {
                print("Overlap \(-overlap.roundedInt) s:")
                print("    Activity \(current.workoutConfiguration.activityType.description): \(current.startDate.timeAndDateText) -> \(current.currentEndDate.timeAndDateText)")
                print("    Activity \(next.workoutConfiguration.activityType.description): \(next.startDate.timeAndDateText) -> \(next.currentEndDate.timeAndDateText)")
            }
            current = next
        }
    }

    convenience init(database: Database) {
        self.init(fileUrl: .init(filePath: "/"), database: database)
    }

    func insert(workout: Workout) throws {
        try workoutsTable.insert(workout)
    }

    func insert(workout: Workout, into store: HKHealthStore) async throws -> HKWorkout? {
        guard let configuration = workout.activities.first?.workoutConfiguration else {
            return nil
        }

        let builder = HKWorkoutBuilder(healthStore: store, configuration: configuration, device: nil)
        return try await builder.finishWorkout()
    }

    func createTables() throws {
        try samples.createAll()
        try workoutsTable.createAll()
        try locationSamples.create(references: dataSeries)
    }
}

private extension HKWorkoutActivity {

    var currentEndDate: Date {
        endDate ?? Date()
    }
}
